// 
//  downloadmanager.js
//  firefox
//  
//  Created by Zak on 2008-06-17.
//  Contributor Brian King
//  Contributor Igor Chornous
//  Copyright 2008-2012 Ant.com. All rights reserved.
// 

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://gre/modules/Downloads.jsm");
Components.utils.import("resource://gre/modules/Task.jsm");
Components.utils.import("resource://gre/modules/FileUtils.jsm");

/**
 * @class AntDownloader
 */
function AntDownloader()
{
    const Cr = Components.results;
    function FileException(errno, innerException) {
        this.getErrno = function() { return errno; };
        this.getInnerException = function() { return innerException; };
    };

    /**
     * Stores video request in the local directory
     *
     * @param videoRequest video source to download from
     * @param doNotOpen if true the function won't open the firefox download manager
     * @return {Boolean} Whether the download has been initiated or not
     */
    this.downloadFlv = function (videoRequest, doNotOpen)
    {
        try {
            let target = setupTargetPath(videoRequest);
            download(videoRequest, target.dest, target.part);
        } catch (e if e instanceof FileException) {
            reportInvalidFolder("AntBar.deniedFolderNotice");
            if (e.getErrno() != Cr.NS_ERROR_FILE_ACCESS_DENIED) {
                AntLib.toLog("Unexpected errno @ downloadFlv: " + e.getErrno());
            }
            return false;
        } catch (e) {
            AntLib.toLog("Unexpected error @ downloadFlv: " + e + "\n" + e.stack);
            reportFailure("");
            return false;
        }
        return true;
    };

    // {{{ notifyDownload private method
    /**
     * @private
     * @member notifyDownload
     * @param {nsIVideoRequest} videoRequest
     * @param {nsIFile} dest
     */
    var notifyDownload = function(videoRequest, dest)
    {
        var videoRequestInfo = {
            name: videoRequest.getName(),
            originUrl: videoRequest.getOrigin().location.href,
            videoUrl: videoRequest.getUri().spec,
            originTitle: AntLib.safeGet(videoRequest.getOrigin(), "title"),
            contentType: videoRequest.getContentType(),
            contentLength: videoRequest.getContentLength()
        };

        AntRPC.download(videoRequestInfo);
        var path = dest.path;
        try {
            AntStorage.beginTransaction();
            AntStorage.addPlaylist('');
            AntStorage.addVideoRecord(
                AntHash.getFileHash(path)
                , videoRequestInfo.name
                , path
                , videoRequestInfo.originUrl
                , videoRequestInfo.videoUrl
                , videoRequestInfo.originTitle
                , 0
                , dest.fileSize
                , ''
                , 0
                , 0
                , (new Date()).getTime());
        } finally {
            try {
                AntStorage.endTransaction();
                var playerWnd = AntBar.getPlayerWnd();
                if (playerWnd)
                    playerWnd.AntPlayer.updateTreeContent();
            }
            catch(e) {
                AntLib.toLog("AntDownloadListener: Download cleanup error: " + e);
            }
        }
    };
    // }}}

    /**
     * Downloads data pointed by {videoRequest} to the local file {dest}.
     * A record is added to the browser's download manager.
     * This method respects the privacy context.
     *
     * @private
     * @member download
     * @param {IVideoRequest} videoRequest
     * @param {nsIFile} dest Download target
     * @param {nsIFile} part File to store partially downloaded content
     */
    var download = function(videoRequest, dest, part)
    {
        try {
            if (!AntLib.inPrivate) {
                notifyDownload(videoRequest, dest);
            }
        }
        catch (ex) {
            AntLib.toLog("Error @download: " + ex);
        }

        Task.spawn(function () {
            let list = yield Downloads.getList(
                AntLib.inPrivate
                    ? Downloads.PRIVATE
                    : Downloads.ALL);

            try {
                let downloadInst = yield Downloads.createDownload({
                    source: {
                        url: videoRequest.getUri().spec,
                        isPrivate: AntLib.inPrivate,
                        referrer: null
                    },
                    target: {
                        path: dest.path,
                        partFilePath: part.path
                    },
                    saver: "copy"
                });
                downloadInst.tryToKeepPartialData = true;

                yield list.add(downloadInst);
                yield downloadInst.start();
            } catch (e) {
                AntLib.toLog("Download error is " + e);
            }
        }).then(null, Components.utils.reportError);
    };

    // {{{ getDestFolder private method
    /**
     * Return the movies destination folder and create it if it doesn't exist
     *
     * @private
     * @member getDestFolder
     * @return {nsIFile} Instance which references the folder
     */
    var getDestFolder = function()
    {
        try {
            let file = new FileUtils.File(AntPrefs.flvDir);

            if (!file.exists())
                file.create(AntLib.CI("nsIFile").DIRECTORY_TYPE
                            , FileUtils.PERMS_DIRECTORY);
            return file;
        } catch (e) {
            throw new FileException(e.result, e);
        }
    };
    // }}}

    // {{{ setupTargetPath private method
    /**
     * @private
     * @member setupTargetPath
     * @param videoRequest flv link
     * @returns {nsIFile} file object for the given flv link
     */
    var setupTargetPath = function(videoRequest)
    {
        let vrName = AntLib.sanitize(videoRequest.getName());
        let vrExt = AntLib.deductExtension(videoRequest.getContentType());
        let fileNameSegment = AntLib.mangleFileName(vrName, vrExt);

        let destFile = getDestFolder();
        destFile.append(fileNameSegment);
        let partFile = destFile.clone();
        partFile.leafName += ".part";

        try {
            destFile.createUnique(
                AntLib.CI("nsIFile").NORMAL_FILE_TYPE
                , FileUtils.PERMS_FILE);
            partFile.createUnique(
                AntLib.CI("nsIFile").NORMAL_FILE_TYPE
                , FileUtils.PERMS_FILE);
            return {'dest': destFile, 
                    'part': partFile};
        } catch (e) {
            AntLib.toLog("Failed to create " + destFile.path + "\nError:\n" + e.result);
            throw new FileException(e.result, e);
        }
    };
    // }}}

    // {{{ reportInvalidFolder private method
    /**
     * @private
     * @member reportInvalidFolder
     * @param {String} id
     */
    var reportInvalidFolder = function(id)
    {
        var acceptButton = new Object();
        acceptButton.label = 'Change';
        acceptButton.accessKey = null;
        acceptButton.popup = null;
        acceptButton.callback = function() { AntBar.doPrefs(); };

        var nb = gBrowser.getNotificationBox();
        nb.appendNotification(
            AntLang.getString(id),
            'ant-denied-folder',
            'chrome://antbar/skin/favico16.png',
            nb.PRIORITY_WARNING_HIGH,
            [acceptButton] );
    };
    // }}}

    // {{{ reportFailure private method
    var reportFailure = function(id)
    {
        var nb = gBrowser.getNotificationBox();
        nb.appendNotification(
            AntLang.getString(id),
            'ant-denied-folder',
            'chrome://antbar/skin/favico16.png',
            nb.PRIORITY_WARNING_HIGH,
            [] );
    };
    // }}}
};





