/**
 * dm.js, 2013
 * @author Igor Chornous ichornous@heliostech.hk
 */
var provs = (function(provs) {

const Ci = Components.interfaces;
const Cc = Components.classes;

const ID_HDSPROTOCOL_CONTRACT = "@mozilla.org/network/protocol;1?name=x-hds";
const ID_IOSERVICE_CONTRACT = "@mozilla.org/network/io-service;1";

Components.utils.import("resource://gre/modules/NetUtil.jsm");

// {{{ DmHdsStream
/**
 * @class DmHdsStream
 * @implements IHdsStream
 * @param {String} name
 */
function DmHdsStream(name)
{
    // {{{ Private fields
    /**
     * @private
     * @field
     * @type Array
     * @name fragments
     */
    var fragments = ["data:binary/octet-stream;base64,RkxWAQUAAAAJAAAAAA=="];

    /**
     * Estimated size of content
     *
     * @private
     * @field
     * @name contentLength
     * @type Number
     */
    var contentLength = undefined;
    // }}}

    /**
     * Public access
     */
    // {{{ addFragment method
    /**
     * @member addFragment
     * @param {String} spec Uri of a fragment
     */
    this.addFragment = function(spec)
    {
        fragments.push(spec);
    };
    // }}}

    // {{{ setContentLength method
    /**
     * @member setContentLength
     * @param {Number} length
     */
    this.setContentLength = function(length)
    {
        contentLength = length;
    };
    // }}}

    // {{{ getName method

    /**
     * @member getName
     * @returns {String}
     */
    this.getName = function()
    {
        return name;
    };

    // }}}

    /**
     * IHdsStream implementation
     */
    // {{{ getContentLength method
    /**
     * @member getContentLength
     * @returns {Number}
     */
    this.getContentLength = function()
    {
        return contentLength;
    };
    // }}}

    // {{{ getFragmentCount method
    /**
     * @member getFragmentCount
     * @returns {Number} Number of fragments in the stream
     */
    this.getFragmentCount = function()
    {
        return fragments.length;
    };
    // }}}

    // {{{ createChannel method
    /**
     * @member createFragmentChannel
     * @param {Number} fragmentIndex
     * @returns {nsIChannel}
     */
    this.createChannel = function(fragmentIndex)
    {
        var ioService = Cc[ID_IOSERVICE_CONTRACT].getService(Ci.nsIIOService);
        var uri = ioService.newURI(
            getFragmentUri(fragmentIndex)
            , null
            , null);
        return ioService.newChannelFromURI(uri);
    };
    // }}}

    /**
     * Aux
     */
    // {{{ getFragmentUri private method
    /**
     * @private
     * @member getFragmentUri
     * @param {Number} index
     * @returns {String} Uri of a fragment
     */
    var getFragmentUri = function(index)
    {
        return fragments[index];
    };
    // }}}
};
// }}}

// {{{ DmVideoRequest

/**
 * @class DmVideoRequest
 * @implements IVideoRequest
 */
function DmVideoRequest()
{
    var ctx = this;

    // {{{ Properties
    /**
     * @name name
     * @type String
     */
    var name = "";

    /**
     * Hds uri
     *
     * @name uri
     * @type nsIURI
     */
    var videoUri = null;

    /**
     * @name origin
     * @type Document
     */
    var origin = null;
    // }}}

    // {{{ init method

    /**
     * @constructor
     * @member init
     * @param {nsIURI} uri
     * @param {DmHdsStream} stream
     * @param {Document} document
     * @returns {undefined} nothing
     */
    this.init = function(uri, stream, document)
    {
        var hdsProtocol = getHdsProtocol();
        var hdsUri = hdsProtocol.createUri(uri);
        hdsProtocol.addStream(hdsUri, stream);
        videoUri = hdsUri;
        origin = document;
        name = AntLib.safeGet(origin, "title") + " [" + stream.getName() + "]";
    };

    // }}}

    /**
     * IVideoRequest implementation
     */
    // {{{ release method
    /**
     * Releases resources associated with the request e.g.
     * in case of x-hds protocol a video stream is removed
     * from the stream collection in the HdsProtocol object
     *
     * @member release
     * @returns {undefined} nothing
     */
    this.release = function()
    {
        var hdsProtocol = getHdsProtocol();
        hdsProtocol.removeStream(videoUri);
        origin = null;
        videoUri = null;
    };
    // }}}

    // {{{ compare method
    /**
     * Checks if both requests are referencing the same video
     *
     * @member compare
     * @param {IVideoRequest} request
     * @returns {Boolean}
     */
    this.compare = function(request)
    {
        var requestPath = request.getUri().spec;
        var ownPath = videoUri.spec;
        return ownPath == requestPath;
    };
    // }}}

    // {{{ getContentType method
    /**
     * @member getContentType
     * @returns {String}
     */
    this.getContentType = function()
    {
        return "video/x-flv";
    };
    // }}}

    // {{{ getContentLength method
    /**
     * @member getContentLength
     * @returns {Number}
     */
    this.getContentLength = function()
    {
        return -1;
    };
    // }}}

    // {{{ getName method
    /**
     * Constructs the human friendly video name
     *
     * @member getDisplayName
     * @returns {String}
     */
    this.getName = function()
    {
        return name;
    };
    // }}}

    // {{{ getOrigin method
    /**
     * Returns a document object associated with the request.
     *
     * @member getOrigin
     * @returns {Document}
     */
    this.getOrigin = function()
    {
        return origin;
    };
    // }}}

    // {{{ getUri method

    /**
     * @member getUri
     * @returns {nsIURI}
     */
    this.getUri = function()
    {
        return videoUri;
    };

    // }}}

    /**
     * Miscellaneous
     */
    // {{{ getHdsProtocol private method
    /**
     * @see @/components/HdsProtocol.js
     * @private
     * @member getHdsProtocol
     * @returns {HdsProtocol}
     */
    var getHdsProtocol = function()
    {
        try
        {
            return Cc[ID_HDSPROTOCOL_CONTRACT]
                .createInstance(Ci.nsIProtocolHandler)
                .wrappedJSObject;
        }
        catch (e)
        {
            AntLib.toLog("x-hds protocol component failed to initialize:\n" +
                         + e.toString());
            throw new Error("Internal failure");
        }
    };
    // }}}
};

// }}}

// {{{ DmSearchResult

/**
 * @class DmSearchResult
 * @implements ISearchResult
 */
function DmSearchResult()
{
    var ctx = this;

    // {{{ Properties

    /**
     * @name manifestUri
     * @type nsIURI
     */
    var manifestUri = null;

    /**
     * @name document
     * @type Document
     */
    var document = null;

    /**
     * @name callback
     * @type Function
     */
    var callback = null;

    // }}}

    /**
     * ISearchResult implementation
     */
    // {{{ asyncFetch method

    /**
     * Asynchronously downloads and parses the manifest
     * @member asyncFetch
     * @param {Function} clbck May be called multiple times.
     *                         An instance of FlvLink is as a single argument
     * @returns {undefined} nothing
     */
    this.asyncFetch = function(clbck)
    {
        callback = clbck;
        addVideoManifestUri(manifestUri);
    };

    // }}}

    /**
     * Internal interface
     */
    // {{{ setManifestUri method
    /**
     * Specifies the video manifest uri. This function must be
     * called prior to the invokation of asyncFetch
     * @member setManifestUri
     * @param {nsIURI} uri
     */
    this.setManifestUri = function(uri)
    {
        manifestUri = uri;
    };
    // }}}

    // {{{ setDocument method
    /**
     * Specifies the document which is associated with the manifest
     * This function must be called prior to the invokation of asyncFetch
     * @member setDocument
     * @param {Document} associatedDocument
     */
    this.setDocument = function(associatedDocument)
    {
        document = associatedDocument;
    };
    // }}}

    /**
     * Implementation
     */
    // {{{ addVideoManifestUri method
    /**
     * Asynchronously downloads a manifest pointed by uri
     *
     * @member addVideoManifestUri
     * @param {nsIURI} uri Uri of the dm's manifest
     */
    var addVideoManifestUri = function(uri)
    {
        withContentUri(uri, ctx.addVideoManifestContent);
    };
    // }}}

    // {{{ addVideoManifestContent method

    /**
     * Synchronously parses the content and builds a valid object of VideoSource
     * @member addVideoManifestContent
     * @param {String} content Dm manifest's content
     */
    this.addVideoManifestContent = function(content, found)
    {
        var manifest = JSON.parse(content);
        if (manifest['version'] != '1')
        {
            // TODO: Log this error
            return;
        }

        var defaultStreamName = manifest['default'];
        for each(var i in manifest['alternates'])
        {
            var name = i['name'];
            var streamManifestUriSpec = i['template'];
            var streamManifestUri = uriFromString(streamManifestUriSpec);
            ctx.addVideoStreamManifestUri(name, streamManifestUri);
        }
    };

    // }}}

    // {{{ addVideoStreamManifestUri method
    /**
     * @member addVideoStreamManifestUri
     * @param {String} name Name of the stream
     * @param {nsIURI} uri Uri of the stream manifest
     */
    this.addVideoStreamManifestUri = function(name, uri)
    {
        withContentUri(
            uri
            , function(content)
            {
                ctx.addVideoStreamManifestContent(uri, name, content);
            });
    };
    // }}}

    // {{{ addVideoStreamManifestContent method

    /**
     * Parses the dm stream manifest and adds a corresponding hds stream to the queue
     * @member addVideoStreamManifestContent
     * @param {nsIURI} uri Uri of the video stream manifest
     * @param {String} name Name of the stream
     * @param {String} content Manifest's content
     */
    this.addVideoStreamManifestContent = function(uri, name, content)
    {
        var streamManifest = JSON.parse(content);
        if (streamManifest['version'] != "1")
        {
            return;
        }

        var baseUri = uriFromString(uri.prePath);
        var hdsStream = new DmHdsStream(name);

        try
        {
            let bitrate = streamManifest['bitrate'];
            let duration = streamManifest['duration'];
            let length = bitrate * duration * 128;
            hdsStream.setContentLength(length);
        }
        catch (e)
        {
            AntLib.toLog("Failed to guess the content length: " + e);
        }

        var j = 1;
        var template = streamManifest['template'];
        for each (var fragment in streamManifest['fragments'])
        {
            for (var i = 0; i < fragment[0]; ++i)
            {
                /**
                 * template contains only the path portion of uri
                 * so we need to resolve it
                 */
                var fragmentUriPathStr = template.replace(
                        /\$fragment\$/i, j.toString());

                var fragmentUriStr = baseUri.resolve(fragmentUriPathStr);
                hdsStream.addFragment(fragmentUriStr);
                ++j;
            }
        }

        addVideoStream(uri, hdsStream);
    };

    // }}}

    // {{{ addVideoStream private method

    /**
     * @private
     * @member addVideoStream
     * @param {nsIURI} uri
     * @param {DmHdsStream} hdsStream
     */
    var addVideoStream = function(uri, hdsStream)
    {
        if (hdsStream.getFragmentCount() == 0)
        {
            AntLib.toLog("Video manifest doesn't contain fragments:\n"
                         + uri.spec + "\n");
            return;
        }

        try
        {
            var videoRequest = new DmVideoRequest();
            videoRequest.init(uri, hdsStream, document);
            callback(videoRequest);
        }
        catch (e)
        {
            AntLib.toLog("Failed to add a stream:\n" + uri.spec
                         + "\nReason: " + e.toString() + "\n");
        }
    };

    // }}}

    // {{{ withContentUri private method

    /**
     * @private
     * @member withContentUri
     * @param {nsIURI} uri Uri of the remote resource
     * @param {Function} func Function to be supplied with content of the resource
     * @param {Function} [err=null] Function to be called in case of failure
     */
    var withContentUri = function(uri, func, err)
    {
        NetUtil.asyncFetch(
            uri
            , function(inputStream, status)
            {
                if (!Components.isSuccessCode(status) && err)
                {
                    err(uri, status);
                    return;
                }

                try
                {
                    var content =
                            NetUtil.readInputStreamToString(
                                inputStream
                                , inputStream.available());
                    func(content);
                }
                catch (e)
                {
                    // TODO: log error
                    return;
                }
            });
    };

    // }}}

    // {{{ uriFromString private method
    /**
     * @private
     * @member uriFromString
     * @param {String} spec
     * @returns {nsIURI}
     */
    var uriFromString = function(spec)
    {
        var ioService = Cc[ID_IOSERVICE_CONTRACT].getService(Ci.nsIIOService);
        return ioService.newURI(spec, null, null);
    };
    // }}}
};

// }}}

// {{{ DmSearchStrategy

/**
 * @class DmSearchStrategy
 * @implements ISearchStrategy
 */
provs.DmSearchStrategy = function()
{
    var ctx = this;

    const reHost = /www\.dailymotion\.com/i;
    const reManifestPath =/\/cdn\/manifest\/video\//i;
    const manifestContentType = "application/vnd.lumberjack.manifest";

    /**
     * ISearchStrategy implementation
     */
    // {{{ isApplicable method

    /**
     * @member isApplicable
     * @param {Document} document
     * @param {nsIHttpChannel} channel
     * @returns {Boolean}
     */
    this.isApplicable = function(document, channel)
    {
        var docUri = document.documentURIObject;
        var reqUri = channel.URI;

        try
        {
            // document's uri could be 'about:blank' and so on
            // in that case the `host accessor will throw an exception
            if (docUri.host.match(reHost))
                return true;
        }
        catch (e) { /** ignore */ }

        try
        {
            if (reqUri.host.match(reHost))
                return true;
        }
        catch (e) { /** ignore */ }

        return false;
    };

    // }}}

    // {{{ search method
    /**
     * @member search
     * @param {Document} document Owning document
     * @param {nsIHttpChannel} channel Request's channel to analyze
     * @param {Function} found The function 'found' is to be called in case if video
     *                         content is found. It may be invoked multiple times.
     *                         The single argument is `flvLink:AntFlvLink
     * @returns {undefined} nothing
     */
    this.search = function(document, channel, found)
    {
        if (!document || !channel || !found)
        {
            // TODO(Igor): Notify error
            return;
        }

        var uri = channel.URI;

        // uri.host & path accessors may throw an exception, but we don't care
        if ((channel.contentType != manifestContentType)
            || (channel.requestMethod != 'GET')
            || !uri.host.match(reHost)
            || !uri.path.match(reManifestPath))
        {
            return;
        }

        var searchResult = new DmSearchResult();
        searchResult.setManifestUri(uri);
        searchResult.setDocument(document);
        searchResult.asyncFetch(found);
    };
    // }}}
};
// }}}

return provs;
})(provs);

