/**
 * yt.js, 2013
 * @author Igor Chornous ichornous@heliostech.hk
 */
var provs = (function(provs) {

const Ci = Components.interfaces;
const Cc = Components.classes;

Components.utils.import("resource://gre/modules/NetUtil.jsm");

// {{{ TracingStreamProxy
/**
 * @class TracingStreamProxy
 * @implements nsIStreamListener
 */
function TracingStreamProxy()
{
    // {{{ setOriginal method
    /**
     * @member setOriginal
     * @param {nsIStreamListener} streamListener
     * @returns {undefined} nothing
     */
    this.setOriginal = function(streamListener)
    {
        originalStreamListener = streamListener;
    };
    // }}}

    // {{{ setOnSuccess method
    /**
     * @member setOnSuccess
     * @param {Function} callback
     * @returns {undefined} nothing
     */
    this.setOnSuccess = function(callback)
    {
        onSuccess = callback;
    };
    // }}}

    // {{{ getContent method
    /**
     * @member getContent
     * @returns {String}
     */
    this.getContent = function()
    {
        return content;
    };
    // }}}

    // {{{ getOrigin method
    /**
     * @private
     * @member getOrigin
     * @returns {Document}
     */
    this.getOrigin = function()
    {
        return origin;
    };
    // }}}

    // {{{ setOrigin method
    /**
     * @private
     * @member setOrigin
     * @param {Document} value
     */
    this.setOrigin = function(value)
    {
        origin = value;
    };
    // }}}

    /**
     * nsIStreamListener implementation
     */
    // {{{ onStartRequest method
    /**
     * @member onStartRequest
     * @see nsIStreamListener#onStartRequest
     */
    this.onStartRequest = function(request, context)
    {
        if (!originalStreamListener)
            return;
        originalStreamListener.onStartRequest(request, context);
    };
    // }}}

    // {{{ onDataAvailable method

    /**
     * @member onDataAvailable
     * @see nsIStreamListener#onDataAvailable
     */
    this.onDataAvailable = function(request, context, inputStream, offset, count)
    {
        if (originalStreamListener)
        {
            originalStreamListener.onDataAvailable(
                request
                , context
                , inputStream
                , offset
                , count);

            var seekableStream = inputStream.QueryInterface(Ci.nsISeekableStream);
            if (!seekableStream)
                return;

            seekableStream.seek(0, 0);
        }

        content = content + NetUtil.readInputStreamToString(inputStream, count);
    };

    // }}}

    // {{{ onStopRequest method

    /**
     * @member onStopRequest
     * @see nsIStreamListener#onStopRequest
     */
    this.onStopRequest = function(aRequest, context, status)
    {
        if (originalStreamListener)
            originalStreamListener.onStopRequest(aRequest, context, status);

        if (!Components.isSuccessCode(status))
            return;

        if (!onSuccess)
            return;

        var document = AntLib.getDocumentByRequest(aRequest);
        origin = document;
        onSuccess();
    };

    // }}}

    // {{{ QueryInterface method

    /**
     * @member QueryInterface
     * @param {Guid} iid Interface identifier
     */
    this.QueryInterface = function(iid)
    {
        if (!iid.equals(Ci.nsISupports) &&
            !iid.equals(Ci.nsIStreamListener))
        {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    };

    // }}}

    /**
     * Object's state
     */
    // {{{ Private properties
    /**
     * Can be accessed via getContent
     *
     * @private
     * @name content
     * @type String
     */
    var content = "";

    /**
     * Can be accessed via getOrigin
     *
     * @private
     * @name origin
     * @type Document
     */
    var origin = null;

    /**
     * @private
     * @name originalStreamListener
     * @type nsIStreamListener
     */
    var originalStreamListener = null;

    /**
     * @private
     * @name onSuccess
     * @type Function
     */
    var onSuccess = null;
    // }}}
};
// }}}

// {{{ YtSearchResult

/**
 * @class YtSearchResult
 * @implements ISearchResult
 */
function YtSearchResult()
{
    var ctx = this;

    const reFmtStreamMap = /url_encoded_fmt_stream_map\":\s*\"([^\"]+)\"/i;

    // {{{ Properties
    /**
     * @private
     * @name content
     * @type String
     */
    var content = null;

    /**
     * @private
     * @name videoOrigin
     * @type Document
     */
    var videoOrigin = null;

    /**
     * @private
     * @name videoNameHint
     * @type String
     */
    var videoNameHint = null;
    // }}}

    // {{{ setContent method
    /**
     * @member setContent
     * @param {String} str
     * @returns {undefined} nothing
     */
    this.setContent = function(str)
    {
        content = str;
    };
    // }}}

    // {{{ setOrigin method
    /**
     * @member setOrigin
     * @param {Document} origin
     * @returns {undefined} nothing
     */
    this.setOrigin = function(origin)
    {
        videoOrigin = origin;
    };
    // }}}

    /**
     * ISearchResult implementation
     */
    // {{{ asyncFetch method

    /**
     * @member asyncFetch
     * @param {Function} clbck May be called multiple times.
     * @returns {undefined} nothing
     */
    this.asyncFetch = function(clbck)
    {
        if (!content || !content.length)
            return;
        if (!videoOrigin)
            return;

        extractVideoInfo();

        var matchResult = content.match(reFmtStreamMap);
        if (!matchResult || (matchResult.length < 2))
            return;

        var fmtStreamMap = matchResult[1];
        var decodedContent = make_unicode_string(fmtStreamMap);
        extractVideoStreams(decodedContent, clbck);
    };

    // }}}

    // {{{ extractVideoStreams private method

    /**
     * @private
     * @member extractVideoStreams
     * @param {String} input
     * @param {Function} callback
     * @returns {undefined} nothing
     */
    var extractVideoStreams = function(input, callback)
    {
        var streams = null;
        try
        {
            streams = parseStreamMap(input);
        }
        catch (e)
        {
            AntLib.toLog("Failed to parse youtube's stream map: " + e.toString());
            return;
        }

        for each (var stream in streams)
        {
            var request = createVideoRequest(stream);
            try
            {
                callback(request);
            }
            catch (e)
            {
                AntLib.toLog("Failed to push a video request:\n"
                             + e.toString() + "\n"
                             + "Stack:\n" + e.stack);
            }
        }
    };

    // }}}

    // {{{ createVideoRequest private method

    /**
     * @private
     * @member createVideoRequest
     * @param {Object} streamInfo
     * @returns {BaseVideoRequest}
     */
    var createVideoRequest = function(streamInfo)
    {
        var videoUri = AntLib.toURI(streamInfo.url);
        var params = {
            uri: videoUri,
            document: videoOrigin,
            contentType: streamInfo.type,
            method: "GET",
            payload: null
        };

        var videoRequest = new provs.BaseVideoRequest();
        videoRequest.init(params);
        videoRequest.setName(getStreamDisplayInfo(streamInfo));
        return videoRequest;
    };

    // }}}

    // {{{ parseStreamMap private method

    /**
     * Parses a video streams descriptor and returns a list of video urls
     *
     * @private
     * @member parseStreamMap
     * @param {String} info Streams descriptor
     * @returns {Array<Object>} Collection of video urls along with some
     *                          additional info. Each video object contains
     *                          following fields: url, url_type and tag
     */
    var parseStreamMap = function(info)
    {
        var result = [];

        var streamCollection = info.split(',');
        for each (var streamInfoStr in streamCollection)
        {
            var streamInfo = decomposeMap(streamInfoStr);
            if (!streamInfo || !streamInfo.url || !streamInfo.itag || !streamInfo.sig)
                continue;

            var url = streamInfo.url + "&signature=" + streamInfo.sig;

            result.push({
                url: url,
                tag: streamInfo.itag,
                type: streamInfo.type,
                quality: streamInfo.quality
            });
        }

        return result;
    };

    // }}}

    // {{{ decomposeMap private method
    /**
     * Given the &-separated string of key-value pairs builds a dictionary
     *
     * @private
     * @name decomposeMap
     * @param {String} str
     * @returns {Map<String,String>}
     */
    var decomposeMap = function(str)
    {
        if (!str || !str.length)
            return null;

        var result = {};

        var streamArgs = str.split('&');
        for each (var streamArg in streamArgs)
        {
            var pair = streamArg.split("=");
            var key = pair[0];
            var value = pair[1];
            result[key] = value ? decodeURIComponent(value) : null;
        }

        return result;
    };
    // }}}

    // {{{ extractVideoInfo private method
    /**
     * @private
     * @member extractVideoInfo
     * @returns {undefined} nothing
     */
    var extractVideoInfo = function()
    {
        var metaElementsCollection = videoOrigin.getElementsByTagName("meta");
        if (!metaElementsCollection || !metaElementsCollection.length)
            return;

        for each (var metaEl in metaElementsCollection)
        {
            if (metaEl.getAttribute("itemprop") == "name")
            {
                videoNameHint = metaEl.getAttribute("content");
                break;
            }
        }

        if (!videoNameHint)
        {
            videoNameHint = AntLib.safeGet(videoOrigin, "title");
        }
    };
    // }}}

    // {{{ getStreamDisplayInfo private method
    /**
     * @private
     * @member getStreamDisplayInfo
     * @param streamInfo
     * @returns {String}
     */
    var getStreamDisplayInfo = function(streamInfo)
    {
        var result = videoNameHint;
        result = result + "[" + streamInfo.quality + "]";
        return result;
    };
    // }}}

    /**
     * Aux
     */
    // {{{ make_unicode_string private method
    /**
     * TODO(Igor): Move this method to AntLib
     *
     * Replaces all occurances of \uxxxx sequences with corresponding unicode characters
     * @private
     * @member make_unicode_string
     * @param {String} str Input
     * @returns {String}
     */
    var make_unicode_string = function(str)
    {
        const reUnicodeEntries = /\\u([0-9a-f]{4})/ig;
        return str.replace(
            reUnicodeEntries
            , function(match, group, offset, str)
            {
                return String.fromCharCode(parseInt(group, 16));
            });
    };
    // }}}
};

// }}}

// {{{ YtSearchStrategy

/**
 * @class YtSearchStrategy
 * @implements ISearchStrategy
 */
provs.YtSearchStrategy = function()
{
    const reHost = /\u0079\u006F\u0075\u0074\u0075\u0062\u0065\.\u0063\u006F\u006D$/i;
    const reVideoPage = /^\/watch/i;

    const domContentLoadedEventName = "DOMContentLoaded";

    /**
     * ISearchStrategy implementation
     */
    // {{{ isApplicable method

    /**
     * @member isApplicable
     * @param {Document} document
     * @param {nsIHttpChannel} channel
     * @returns {Boolean}
     */
    this.isApplicable = function(document, channel)
    {
        var docUri = document.documentURIObject;
        var reqUri = channel.URI;

        try
        {
            // document's uri could be 'about:blank' and so on
            // in that case the `host accessor will throw an exception
            if (docUri.host.match(reHost))
                return true;
        }
        catch (e) { /** ignore */ }

        try
        {
            if (reqUri.host.match(reHost))
                return true;
        }
        catch (e) { /** ignore */ }

        return false;
    };

    // }}}

    // {{{ search method
    /**
     * @member search
     * @param {Document} document Owning document
     * @param {nsIHttpChannel} channel Request's channel to analyze
     * @param {Function} found  See {AntGrabber#foundFlvLink} for details
     * @returns {undefined} nothing
     */
    this.search = function(document, channel, found)
    {
        if (!document || !channel || !found)
        {
            // TODO(Igor): Notify error
            return;
        }

        var requestUri = channel.URI;
        // requestUri.host & path accessors may throw an exception, but we don't care
        if (!requestUri.host.match(reHost)
            || !requestUri.path.match(reVideoPage)
            || (channel.contentType != 'text/html')
            || (channel.requestMethod != 'GET'))
            return;

        parseVideoStreamsAsync(channel, found);
    };
    // }}}

    // {{{ parseVideoStreamsAsync private method

    /**
     * Installs a special callback which will get called when user has completed
     * downloading data from the channel's stream. The callback is fed up
     * with downloaded data. Once called it searches for video urls
     *
     * @private
     * @member parseVideoStreamsAsync
     * @param {nsIHttpChannel} channel
     * @param {Function} found
     */
    var parseVideoStreamsAsync = function(channel, found)
    {
        var traceableChannel = channel.QueryInterface(Ci.nsITraceableChannel);
        if (!traceableChannel)
        {
            AntLib.toLog("Unable to trace the channel\n");
            return;
        }

        var bodyListener = new TracingStreamProxy();
        bodyListener.setOnSuccess(
            function()
            {
                scheduleParseContent(bodyListener, found);
            });

        bodyListener.setOriginal(
            traceableChannel.setNewListener(bodyListener));
    };

    // }}}

    // {{{ trackPageMutations private method
    /**
     * Attachs a mutation observer to certain nodes in the document.
     * When a node changes the onChange callback is to be invoked.
     *
     * @private
     * @member trackPageMutations
     * @param {Document} doc Document to track
     * @param {Function} onChange To be invoked when document significantly changes
     * @returns {Boolean} Success
     */
    var trackPageMutations = function(doc, onChange)
    {
        if (!onChange)
            throw new Error("onChange is mandatory");
        if (!doc)
            throw new Error("doc is mandatory");

        var target = doc.getElementById("player");
        if (!target)
            return false;

        var url = doc.URL;
        var observer = new MutationObserver(
            function(mutations) {
                if (url == doc.URL)
                    return;

                url = doc.URL;
                try {
                    onChange();
                } catch (e) {
                    AntLib.toLog("Unexpected exception @ mutation observer: "
                                 + e + "\n");
                }
            });

        var config = { attributes: true, childList: true, characterData: true };
        observer.observe(target, config);
        return true;
    };
    // }}}

    // {{{ downloadContent private method
    /**
     * @private
     * @member downloadContent
     * @param {Document} origin
     * @param {String} url
     * @param {Function} clbck
     */
    var downloadContentAsync = function(origin, url, clbck)
    {
        var listener = new TracingStreamProxy();
        listener.setOnSuccess(
            function()
            {
                listener.setOrigin(origin);
                clbck(listener);
            });
        var channel = AntLib.createChannelFromURL(url);
        channel.notificationCallbacks = listener;
        channel.asyncOpen(listener, null);
    };
    // }}}

    // {{{ scheduleParseContent private method
    /**
     * Starts to parse the page content as soon as the document
     * fires the onLoad event. Thereby we are able to access all
     * the information via document's accessors
     *
     * @private
     * @member scheduleParseContent
     * @param {TracingStreamProxy} bodyListener
     * @param {Function} found
     * @returns {undefined} nothing
     */
    var scheduleParseContent = function(bodyListener, found)
    {
        var origin = bodyListener.getOrigin();

        var onNewPushStateDocument = function()
        {
            try {
                // This callback will clear the video list on the first call
                var callback = (function() {
                    var ft = true;
                    return function(request) {
                        found(request, ft);
                        ft = false;
                    };
                })();

                downloadContentAsync(
                    origin
                    , origin.URL
                    , function(listener) {
                        parseVideoPageContent(listener, callback);
                    });
            } catch (e) {
                AntLib.toLog("Unexpected exception in callback:\n" + e.toString()
                             + "\nStack:\n" + e.stack + "\n");
            }
        };

        var onLoadCompleted = function()
        {
            try
            {
                origin.removeEventListener(
                    domContentLoadedEventName
                    , onLoadCompleted);

                parseVideoPageContent(bodyListener, found);
                if (!trackPageMutations(origin, onNewPushStateDocument))
                {
                    AntLib.toLog("Failed to install mutation observer\n");
                }
            }
            catch (e)
            {
                AntLib.toLog("Unexpected exception in callback:\n" + e.toString()
                             + "\nStack:\n" + e.stack + "\n");
            }
        };

        origin.addEventListener(
            domContentLoadedEventName
            , onLoadCompleted);
    };
    // }}}

    // {{{ parseVideoPageContent private method
    /**
     * @private
     * @member parseVideoPageContent
     * @param {TracingStreamProxy} contentSource
     * @param {Function} found
     * @returns {undefined} nothing
     */
    var parseVideoPageContent = function(contentSource, found)
    {
        try
        {
            var origin = contentSource.getOrigin();
            var content = contentSource.getContent();
            if (!origin || !content || !content.length)
                return;

            var result = new YtSearchResult();
            result.setContent(content);
            result.setOrigin(origin);
            result.asyncFetch(found);
        }
        catch (e)
        {
            AntLib.toLog("Unexpected exception:\n"
                         + e.toString()
                         + "\nStack:\n" + e.stack + "\n");
        }
    };
    // }}}
};

// }}}

return provs;
})(provs);
